;*=====================================================================*/
;*    serrano/prgm/project/bigloo/runtime/Eval/evmeaning.scm           */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Wed Aug  4 10:48:41 1993                          */
;*    Last change :  Mon Apr 12 21:09:12 2004 (serrano)                */
;*    -------------------------------------------------------------    */
;*    L'interprete de bigloo (la version `lambda')                     */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module __evmeaning
   
   (include "Eval/byte-code.sch")
   
   (import  __type
	    __error
	    __bigloo
	    __tvector
	    __structure
	    __tvector
	    __bexit
	    __os
	    __bit
	    
	    __r4_numbers_6_5
	    __r4_numbers_6_5_fixnum
	    __r4_numbers_6_5_flonum
	    __r4_characters_6_6
	    __r4_equivalence_6_2
	    __r4_booleans_6_1
	    __r4_symbols_6_4
	    __r4_strings_6_7
	    __r4_pairs_and_lists_6_3
	    __r4_input_6_10_2
	    __r4_control_features_6_9
	    __r4_vectors_6_8
	    __r4_ports_6_10_1
	    __r4_output_6_10_3
	    
	    __progn
	    __evenv
	    __evcompile)
   
   (static  (update-eval-global! variable value))
   
   (export  (evmeaning exp env)
	    (evmeaning-reset-error!)
	    (evmeaning-notify-error proc mes obj)
	    (evmeaning-location))
   
   (extern  (%funcall-0::obj  (::procedure)
			      "eval_funcall_0")
	    (%funcall-1::obj  (::procedure ::obj)
			      "eval_funcall_1")
	    (%funcall-2::obj  (::procedure ::obj ::obj)
			      "eval_funcall_2")
	    (%funcall-3::obj  (::procedure ::obj ::obj ::obj)
			      "eval_funcall_3")
	    (%funcall-4::obj  (::procedure ::obj ::obj ::obj ::obj)
			      "eval_funcall_4")
	    (%eval-apply::obj (::procedure ::obj)
			      "eval_apply"))
   
   (java    (class foreign
	       (method static %funcall-0::obj (::procedure)
		       "eval_funcall_0")
	       (method static %funcall-1::obj (::procedure ::obj)
		       "eval_funcall_1")
	       (method static %funcall-2::obj (::procedure ::obj ::obj)
		       "eval_funcall_2")
	       (method static %funcall-3::obj (::procedure ::obj ::obj ::obj)
		       "eval_funcall_3")
	       (method static %funcall-4::obj (::procedure ::obj ::obj ::obj ::obj)
		       "eval_funcall_4")
	       (method static %eval-apply::obj (::procedure ::obj)
		       "eval_apply"))))

;*---------------------------------------------------------------------*/
;*    evmeaning ...                                                    */
;*---------------------------------------------------------------------*/
(define (evmeaning code stack)
   (if (evcode? code)
       (begin
	  (set! *current-bcode* code)
 	  (case (evcode-op code)
	     ;; les erreurs
	     ((-2)
	      (apply evmeaning-error code (evcode-ref code 0)))
	     ;; La seule constante qui nessecite un codage: les `vecteurs'
	     ((-1)
	      (evcode-ref code 0))
	     ;; la premiere variable locale
	     ((0)   
	      (car stack))
	     ;; la deuxieme variable locale
	     ((1)   
	      (cadr stack))
	     ;; la troisieme variable locale
	     ((2)   
	      (caddr stack))
	     ;; la quatrieme variable locale
	     ((3)   
	      (cadddr stack))
	     ;; les variables locales plus profondes
	     ((4)   
	      (let ((offset (evcode-ref code 0)))
		 (do ((i 4 (+fx i 1))
		      (env (cddddr stack) (cdr env)))
		     ((=fx i offset) (car env)))))
	     ;; la reference d'une variable globale mutable
	     ((5)
	      (__evmeaning_address-ref
	       (eval-global-value (evcode-ref code 0))))
	     ;; la reference d'une variable globale non mutable
	     ((6)
	      (eval-global-value (evcode-ref code 0)))
	     ;; la reference a une variable par encore definie
	     ((7)
	      (let* ((name (evcode-ref code 0))
		     (global (eval-lookup name)))
		 (if (eval-global? global)
		     (begin
			;; we change the value of the byte-code
			;; because, now, the variable is bound
			(evcode-op-set! code 6)
			(evcode-set! code 0 global)
			(eval-global-value global))
		     (evmeaning-error code
				      "eval"
				      "Unbound variable"
				      name))))
	     ;; la mutation d'une variable globale
	     ((8)
	      (let ((var (evcode-ref code 0))
		    (val (evmeaning (evcode-ref code 1) stack)))
		 (update-eval-global! var val)
		 (unspecified)))
	     ((9)
	      (let* ((name (evcode-ref code 0))
		     (value (evcode-ref code 1))
		     (global (eval-lookup name)))
		 (if (eval-global? global)
		     (begin
			(evcode-op-set! code 8)
			(evcode-set! code 0 global)
			(evcode-set! code 1 value)
			(evmeaning code stack))
		     (evmeaning-error code
				      "eval"
				      "Unbound variable"
				      name))))
	     ;; la mutation de la premiere variable locale
	     ((10)
	      (set-car! stack (evmeaning (evcode-ref code 0) stack))
	      (unspecified))
	     ;; la mutation de la deuxieme variable locale
	     ((11)
	      (set-car! (cdr stack) (evmeaning (evcode-ref code 0) stack))
	      (unspecified))
	     ;; la mutation de la troisieme variable locale
	     ((12)
	      (set-car! (cddr stack) (evmeaning (evcode-ref code 0) stack))
	      (unspecified))
	     ;; la mutation de la quatrieme variable locale
	     ((13)
	      (set-car! (cdddr stack) (evmeaning (evcode-ref code 0) stack))
	      (unspecified))
	     ;; la mutation des variables locales profondes
	     ((14)
	      (let ((offset (evcode-ref code 0))
		    (value  (evmeaning (evcode-ref code 1) stack)))
		 (do ((i 4 (+fx i 1))
		      (env (cddddr stack) (cdr env)))
		     ((=fx i offset) (set-car! env value)))
		 (unspecified)))
	     ;; la conditionnelle
	     ((15) 
	      (if (evmeaning (evcode-ref code 0) stack)
		  (evmeaning (evcode-ref code 1) stack)
		  (evmeaning (evcode-ref code 2) stack)))
	     ((67)
	      ;; or
	      (let ((len (evcode-length code)))
		 (let loop ((i 0))
		    (if (<fx i len)
			(or (evmeaning (evcode-ref code i) stack)
			    (loop (+fx i 1)))
			#f))))
	     ((68)
	      ;; and
	      (let ((len (evcode-length code)))
		 (let loop ((i 0)
			    (l #t))
		    (if (<fx i len)
			(let ((l (evmeaning (evcode-ref code i) stack)))
			   (and l (loop (+fx i 1) l)))
			l))))
	     ;; sequence (by construction, the length is at least 1)
	     ((16)
	      (let ((len (-fx (evcode-length code) 1)))
		 (let loop ((i 0))
		    (if (=fx i len)
			(evmeaning (evcode-ref code i) stack)
			(begin
			   (evmeaning (evcode-ref code i) stack)
			   (loop (+fx i 1)))))))
	     ;; la forme define (globale) lambda
	     ((17)
	      (let ((var (evcode-ref code 0))
		    (val (evcode-ref code 1)))
		 (let ((cell (eval-lookup var)))
		    (if (eval-global? cell)
			(begin
			   (evmeaning-warning code
					      "eval"
					      #\Newline
					      "redefinition of variable -- "
					      var)
			   (update-eval-global! cell
						(evmeaning (force val) '())))
			(let ((cell (vector 0 var (unspecified))))
 			   (bind-eval-global! var cell)
			   ;; on le fait en deux fois pour etre
			   ;; sur que la liaison existe.
			   (let ((value (evmeaning (force val) '())))
			      (set-eval-global-value! cell value))))
		    var)))
	     ;; la forme define (globale) value
	     ((63)
	      (let* ((var (evcode-ref code 0))
		     (val (evmeaning (evcode-ref code 1) '())))
		 (let ((cell (eval-lookup var)))
		    (if (eval-global? cell)
			(begin
			   (evmeaning-warning code
					      "eval"
					      #\Newline
					      "redefinition of variable -- "
					      var)
			   (update-eval-global! cell val))
			(let ((cell (vector 0 var (unspecified))))
			   (bind-eval-global! var cell)
			   ;; on le fait en deux fois pour etre
			   ;; sur que la liaison existe.
			   (set-eval-global-value! cell val)))
		    var)))
	     ;; bind-exit
	     ((18)
	      (bind-exit (__dummy__)
		 ((evmeaning (evcode-ref code 0) stack) __dummy__)))
	     ;; unwind-protect
	     ((64)
	      (let ((body (evcode-ref code 0))
		    (protect (evcode-ref code 1)))
		 (unwind-protect (evmeaning body stack)
				 (evmeaning protect stack))))
	     ;; l'appel de fonction d'arite 0
	     ((19)
	      (let* ((name (evcode-ref code 0))
		     (fun (__evmeaning_address-ref
			   (eval-global-value (evcode-ref code 1)))))
		 (funcall-0 name fun)))
	     ((57)
	      (let* ((name (evcode-ref code 0))
		     (fun (eval-global-value (evcode-ref code 1))))
		 (funcall-0 name fun)))
	     ;; l'appel de fonction d'arite 1
	     ((20)
	      (let* ((name (evcode-ref code 0))
		     (fun (__evmeaning_address-ref
			   (eval-global-value (evcode-ref code 1))))
		     (a0 (evmeaning (evcode-ref code 2) stack)))
		 (set! *current-bcode* code)
		 (funcall-1 name fun a0)))
	     ((58)
	      (let* ((name (evcode-ref code 0))
		     (fun (eval-global-value (evcode-ref code 1)))
		     (a0 (evmeaning (evcode-ref code 2) stack)))
		 (set! *current-bcode* code)
		 (funcall-1 name fun a0)))
	     ;; l'appel de fonction d'arite 2
	     ((21)
	      (let* ((name (evcode-ref code 0))
		     (fun (__evmeaning_address-ref
			   (eval-global-value (evcode-ref code 1))))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack)))
		 (set! *current-bcode* code)
		 (funcall-2 name fun a0 a1)))
	     ((59)
	      (let* ((name (evcode-ref code 0))
		     (fun (eval-global-value (evcode-ref code 1)))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack)))
		 (set! *current-bcode* code)
		 (funcall-2 name fun a0 a1)))
	     ;; l'appel de fonction d'arite 3
	     ((22)
	      (let* ((name (evcode-ref code 0))
		     (fun (__evmeaning_address-ref
			   (eval-global-value (evcode-ref code 1))))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack))
		     (a2 (evmeaning (evcode-ref code 4) stack)))
		 (set! *current-bcode* code)
		 (funcall-3 name fun a0 a1 a2)))
	     ((60)
	      (let* ((name (evcode-ref code 0))
		     (fun (eval-global-value (evcode-ref code 1)))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack))
		     (a2 (evmeaning (evcode-ref code 4) stack)))
		 (set! *current-bcode* code)
		 (funcall-3 name fun a0 a1 a2)))
	     ;; l'appel de fonction d'arite 4
	     ((23)
	      (let* ((name (evcode-ref code 0))
		     (fun (__evmeaning_address-ref
			   (eval-global-value (evcode-ref code 1))))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack))
		     (a2 (evmeaning (evcode-ref code 4) stack))
		     (a3 (evmeaning (evcode-ref code 5) stack)))
		 (set! *current-bcode* code)
		 (funcall-4 name fun a0 a1 a2 a3)))
	     ((61)
	      (let* ((name (evcode-ref code 0))
		     (fun (eval-global-value (evcode-ref code 1)))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack))
		     (a2 (evmeaning (evcode-ref code 4) stack))
		     (a3 (evmeaning (evcode-ref code 5) stack)))
		 (set! *current-bcode* code)
		 (funcall-4 name fun a0 a1 a2 a3)))
	     ;; l'appel de fonction de plus de quatre arguments
	     ((24)
	      (let* ((name (evcode-ref code 0))
		     (fun (__evmeaning_address-ref
			   (eval-global-value (evcode-ref code 1))))
		     (args  (evcode-ref code 2)))
		 (let loop ((args args)
			    (res  '())
			    (len   0))
		    (if (null? args)
			(begin
			   (set! *current-bcode* code)
			   (eval-apply name fun len (reverse! res)))
			(loop (cdr args)
			      (cons (evmeaning (car args) stack)
				    res)
			      (+fx 1 len))))))
	     ((62)
	      (let* ((name (evcode-ref code 0))
		     (fun (eval-global-value (evcode-ref code 1)))
		     (args (evcode-ref code 2)))
		 (let loop ((args args)
			    (res '())
			    (len  0))
		    (if (null? args)
			(begin
			   (set! *current-bcode* code)
			   (eval-apply name fun len (reverse! res)))
			(loop (cdr args)
			      (cons (evmeaning (car args) stack)
				    res)
			      (+fx len 1))))))
	     ;; l'appel de fonction de compilee anonyme d'arite 0
	     ((25)
	      ((evcode-ref code 0)))
	     ;; l'appel de fonction de compilee anonyme d'arite 1
	     ((26)
	      (let* ((fun (evcode-ref code 0)))
		 (set! *current-bcode* code)
		 (fun (evmeaning (evcode-ref code 1) stack))))
	     ;; l'appel de fonction de compilee anonyme d'arite 2
	     ((27)
	      (let* ((fun (evcode-ref code 0))
		     (a0 (evmeaning (evcode-ref code 1) stack))
		     (a1 (evmeaning (evcode-ref code 2) stack)))
		 (set! *current-bcode* code)
		 (fun a0 a1)))
	     ;; l'appel de fonction de compilee anonyme d'arite 3
	     ((28)
	      (let* ((fun (evcode-ref code 0))
		     (a0 (evmeaning (evcode-ref code 1) stack))
		     (a1 (evmeaning (evcode-ref code 2) stack))
		     (a2 (evmeaning (evcode-ref code 3) stack)))
		 (set! *current-bcode* code)
		 (fun a0 a1 a2)))
	     ;; l'appel de fonction de compilee anonyme d'arite 4
	     ((29)
	      (let* ((fun (evcode-ref code 0))
		     (a0 (evmeaning (evcode-ref code 1) stack))
		     (a1 (evmeaning (evcode-ref code 2) stack))
		     (a2 (evmeaning (evcode-ref code 3) stack))
		     (a3 (evmeaning (evcode-ref code 4) stack)))
		 (set! *current-bcode* code)
		 (fun a0 a1 a2 a3)))
	     ;; l'appel de fonction de compilee anonyme d'arite plus que 4
	     ((30)
	      (let ((eargs (map (lambda (x)
				   (evmeaning x stack))
				(evcode-ref code 1))))
		 (set! *current-bcode* code)
		 (apply (evcode-ref code 0) eargs)))
	     ;; l'application classique d'arite 0
	     ((31)
	      (let* ((name (evcode-ref code 0))
		     (fun (evmeaning (evcode-ref code 1) stack)))
		 (set! *current-bcode* code)
		 (funcall-0 name fun)))
	     ;; l'application classique d'arite 1
	     ((32)
	      (let* ((name (evcode-ref code 0))
		     (fun (evmeaning (evcode-ref code 1) stack))
		     (a0 (evmeaning (evcode-ref code 2) stack)))
		 (set! *current-bcode* code)
		 (funcall-1 name fun a0)))
	     ;; l'application classique d'arite 2
	     ((33)
	      (let* ((name (evcode-ref code 0))
		     (fun (evmeaning (evcode-ref code 1) stack))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack)))
		 (set! *current-bcode* code)
		 (funcall-2 name fun a0 a1)))
	     ;; l'application classique d'arite 3
	     ((34)
	      (let* ((name (evcode-ref code 0))
		     (fun (evmeaning (evcode-ref code 1) stack))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack))
		     (a2 (evmeaning (evcode-ref code 4) stack)))
		 (set! *current-bcode* code)
		 (funcall-3 name fun a0 a1 a2)))
	     ;; l'application classique d'arite 4
	     ((35)
	      (let* ((name (evcode-ref code 0))
		     (fun (evmeaning (evcode-ref code 1) stack))
		     (a0 (evmeaning (evcode-ref code 2) stack))
		     (a1 (evmeaning (evcode-ref code 3) stack))
		     (a2 (evmeaning (evcode-ref code 4) stack))
		     (a3 (evmeaning (evcode-ref code 5) stack)))
		 (set! *current-bcode* code)
		 (funcall-4 name fun a0 a1 a2 a3)))
	     ;; l'application classique de plus de 4 arguments
	     ((36)
	      (let* ((name (evcode-ref code 0))
		     (fun (evmeaning (evcode-ref code 1) stack)))
 		 (let loop ((args (evcode-ref code 2))
			    (new '())
			    (len 0))
		    (if (null? args)
			(begin
			   (set! *current-bcode* code)
			   (eval-apply name fun len (reverse! new)))
			(loop (cdr args)
			      (cons (evmeaning (car args) stack) new)
			      (+fx 1 len))))))

	     ;; les lambda tracees d'arite 0
	     ((37)
	      (let ((body (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda ()
		    (let ((where where))
		       (let ()
			  ;; dont try to change something to this code
			  ;; in particular don't remove the binding
			  ;; (where where) nor the side effect
			  ;; (set! where where), they are use to allow
			  ;; the compilation of this handly introduced
			  ;; push-trace form (the same hack is used everywhere
			  ;; in this module).
	 		  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning body stack)))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda tracees d'arite 1
	     ((38)
	      (let ((body  (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda (x)
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning body (cons x stack))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda tracees d'arite 2
	     ((39)
	      (let ((body  (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda (x y)
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning body (cons x (cons y stack)))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda tracees d'arite 3
	     ((40)
	      (let ((body  (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda (x y z)
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning body
						(cons x
						      (cons y
							    (cons z stack))))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda tracees d'arite 4
	     ((41)
	      (let ((body  (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda (x y z t)
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning
				      body
				      (cons x
					    (cons y
						  (cons z
							(cons t
							      stack)))))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda non tracees d'arite 0
	     ((42)
	      (lambda ()
		 (evmeaning (evcode-ref code 0)
			    stack)))
	     ;; les lambda non tracees d'arite 1
	     ((43)
	      (lambda (x)
		 (evmeaning (evcode-ref code 0)
			    (cons x stack))))
	     ;; les lambda non tracees d'arite 2
	     ((44)
	      (lambda (x y)
		 (evmeaning (evcode-ref code 0)
			    (cons x (cons y stack)))))
	     ;; les lambda non tracees d'arite 3
	     ((45)
	      (lambda (x y z)
		 (evmeaning (evcode-ref code 0)
			    (cons x (cons y (cons z stack))))))
	     ;; les lambda non tracees d'arite 4
	     ((46)
	      (lambda (x y z t)
		 (evmeaning (evcode-ref code 0)
			    (cons x (cons y (cons z (cons t stack)))))))
	     ;; les lambda tracees d'arite -1
	     ((47)
	      (let ((body (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda x
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning body (cons x stack))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda tracees d'arite -2
	     ((48)
	      (let ((body  (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda (x . y)
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning body (cons x (cons y stack)))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda tracees d'arite -3
	     ((49)
	      (let ((body (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda (x y . z)
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning body
						(cons x
						      (cons y
							    (cons z stack))))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda tracees d'arite -4
	     ((50)
	      (let ((body  (evcode-ref code 0))
		    (where (evcode-ref code 1)))
		 (lambda (x y z . t)
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((res (evmeaning
				      body
				      (cons x
					    (cons y
						  (cons z
							(cons t
							      stack)))))))
			     (c-pop-trace)
			     res))))))
	     ;; les lambda non tracees d'arite -1
	     ((51)
	      (lambda x
		 (evmeaning (evcode-ref code 0)
			    (cons x stack))))
	     ;; les lambda non tracees d'arite -2
	     ((52)
	      (lambda (x . y)
		 (evmeaning (evcode-ref code 0)
			    (cons x (cons y stack)))))
	     ;; les lambda non tracees d'arite -3
	     ((53)
	      (lambda (x y . z)
		 (evmeaning (evcode-ref code 0)
			    (cons x (cons y (cons z stack))))))
	     ;; les lambda non tracees d'arite -4
	     ((54)
	      (lambda (x y z . t)
		 (evmeaning (evcode-ref code 0)
			    (cons x (cons y (cons z (cons t stack)))))))
	     ;; les lambdas de plus de 4 arguments tracees
	     ((55)
	      (let ((body (evcode-ref code 0))
		    (where (evcode-ref code 1))
		    (formals (evcode-ref code 2)))
		 (lambda x
		    (let ((where where))
		       (let ()
			  (c-push-trace where)
			  (set! where where)
			  (let ((env2 (let _loop_ ((actuals x)
						   (formals formals))
					 (cond
					    ((null? formals)
					     (if (not (null? actuals))
						 (evmeaning-error
						  code
						  "eval"
						  "Too many arguments provided"
						  actuals)
						 stack))
					    ((not (pair? formals))
					     (cons actuals stack))
					    ((null? actuals)
					     (evmeaning-error
					      code
					      "eval"
					      "Too few arguments provided"
					      formals))
					    (else
					     (cons (car actuals)
						   (_loop_ (cdr actuals)
							   (cdr formals))))))))
			     (let ((res (evmeaning body env2)))
				(c-pop-trace)
				res)))))))
	     ;; les lambdas de plus de 4 arguments non tracees
	     ((56)
	      (let ((body (evcode-ref code 0))
		    (formals (evcode-ref code 1)))
		 (lambda x
		    (let ((env2 (let _loop_ ((actuals x)
					     (formals formals))
				   (cond
				      ((null? formals)
				       (if (not (null? actuals))
					   (evmeaning-error
					    code
					    "eval"
					    "Too many arguments provided"
					    actuals)
					   stack))
				      ((not (pair? formals))
				       (cons actuals stack))
				      ((null? actuals)
				       (evmeaning-error
					code
					"eval"
					"Too few arguments provided"
					formals))
				      (else
				       (cons (car actuals)
					     (_loop_ (cdr actuals)
						     (cdr formals))))))))
		       (evmeaning body env2)))))
	     ((65)
	      ;; let (bindings are stored reversed!)
	      (let loop ((vals (evcode-ref code 1))
			 (env stack))
		 (if (null? vals)
		     (evmeaning (evcode-ref code 0) env)
		     (loop (cdr vals)
			   (cons (evmeaning (car vals) stack) env)))))
	     ((66)
	      ;; let*
	      (let loop ((vals (evcode-ref code 1))
			 (env stack))
		 (if (null? vals)
		     (evmeaning (evcode-ref code 0) env)
		     (loop (cdr vals)
			   (cons (evmeaning (car vals) env) env)))))
	     ((70)
	      ;; letrec
	      (let* ((vals (evcode-ref code 1))
		    (env2 (append (make-list (length vals)) stack)))
		 (let loop ((vals vals)
			    (env3 env2))
		    (if (null? vals)
			(evmeaning (evcode-ref code 0) env2)
			(begin
			   (set-car! env3 (evmeaning (car vals) env2))
			   (loop (cdr vals) (cdr env3)))))))
	     ((69)
	      ;; do (increments are encoded in the body)
	      (let loop ((vals (evcode-ref code 0))
			 (env stack))
		 (if (null? vals)
		     (let ((test (evcode-ref code 1))
			   (endv (evcode-ref code 2))
			   (body (evcode-ref code 3)))
			(let loop ()
			   (if (evmeaning test env)
			       (evmeaning endv env)
			       (begin
				  (evmeaning body env)
				  (loop)))))
		     (loop (cdr vals)
			   (cons (evmeaning (car vals) stack) env)))))
	     (else
	      ;; unknown byte code
	      (evmeaning-error code
			       "evmeaning (internal error)"
			       "unknown byte-code"
			       code))))
       code))

;*---------------------------------------------------------------------*/
;*    evmeaning-error ...                                              */
;*---------------------------------------------------------------------*/
(define (evmeaning-error bcode proc mes obj)
   (if (evcode? bcode)
       (match-case (evcode-loc bcode)
	  ((at ?fname ?loc ?line)
	   (error/location proc mes obj fname loc))
	  (else
	   (error proc mes obj)))
       (error proc mes obj)))

;*---------------------------------------------------------------------*/
;*    evmeaning-warning ...                                            */
;*---------------------------------------------------------------------*/
(define (evmeaning-warning bcode . args)
   (if (evcode? bcode)
       (match-case (evcode-loc bcode)
	  ((at ?fname ?loc ?line)
	   (apply warning/location `(,fname ,loc ,@args)))
	  (else
	   (apply warning args)))
       (apply warning args)))

;*---------------------------------------------------------------------*/
;*    evmeaning-notify-error ...                                       */
;*---------------------------------------------------------------------*/
(define (evmeaning-notify-error proc mes obj)
   (define (byte-code-error)
      (if (evcode? *current-bcode*)
	  (match-case (evcode-loc *current-bcode*)
	     ((at ?fname ?loc ?line)
	      (set! *current-bcode* #f)
	      (let ((err (error/location-file fname loc)))
		 (if (procedure? err)
		     (err proc mes obj)
		     (notify-error proc mes obj))))
	     (else
	      (notify-error proc mes obj)))
	  (notify-error proc mes obj)))
   (if (epair? obj)
       (match-case (cer obj)
	  ((at ?fname ?loc ?line)
	   (set! *current-bcode* #f)
	   (let ((err (error/location-file fname loc)))
	      (if (procedure? err)
		  (err proc mes obj)
		  (notify-error proc mes obj))))
	  (else
	   (byte-code-error)))
       (byte-code-error)))

;*---------------------------------------------------------------------*/
;*    evmeaning-location ...                                           */
;*---------------------------------------------------------------------*/
(define (evmeaning-location)
   (if (evcode? *current-bcode*)
       (let ((p (evcode-loc *current-bcode*)))
	  (match-case p
	     ((at ?fname ?loc ?line)
	      p)
	     (else
	      #f)))
       #f))
   
;*---------------------------------------------------------------------*/
;*    evmeaning-funcall-location ...                                   */
;*---------------------------------------------------------------------*/
(define (evmeaning-funcall-location o)
   (let ((bcode (list-ref *funcall-bcodes*
			  (- *funcall-bcode-length*
			     (modulo o *funcall-bcode-length*)))))
      (if (evcode? bcode)
	  (let ((p (evcode-loc bcode)))
	     (match-case p
		((at ?fname ?loc ?line)
		 p)
		(else
		 #f)))
	  #f)))
   
;*---------------------------------------------------------------------*/
;*    evmeaning-reset-error! ...                                       */
;*---------------------------------------------------------------------*/
(define (evmeaning-reset-error!)
   (set! *current-bcode* #f))
   
;*---------------------------------------------------------------------*/
;*      update-eval-global! ...                                        */
;*---------------------------------------------------------------------*/
(define (update-eval-global! variable val)
   (if (eq? (eval-global-tag variable) 1)
       (__evmeaning_address-set! (eval-global-value variable) val)
       (set-eval-global-value! variable val))
   (eval-global-name variable))

;*---------------------------------------------------------------------*/
;*    *current-bcode* ...                                              */
;*---------------------------------------------------------------------*/
(define *current-bcode* #f)

;*---------------------------------------------------------------------*/
;*    *funcall-bcodes* ...                                             */
;*---------------------------------------------------------------------*/
(define *funcall-bcode-length*
   (let ((std (getenv "BIGLOOSTACKDEPTH")))
      (if (number? std)
	  (max std 10)
	  10)))
(define *funcall-bcodes*
   (let ((l (make-list *funcall-bcode-length* #f)))
      (set-cdr! (last-pair l) l)
      l))

;*---------------------------------------------------------------------*/
;*    funcall-0 ...                                                    */
;*---------------------------------------------------------------------*/
(define-inline (funcall-0 name fun)
   (cond
      ((not (procedure? fun))
       (error "eval" "Not a procedure" name))
      ((not (correct-arity? fun 0))
       (error "eval" "Wrong number of argument" name))
      (else
       (%funcall-0 fun))))

;*---------------------------------------------------------------------*/
;*    funcall-1 ...                                                    */
;*---------------------------------------------------------------------*/
(define-inline (funcall-1 name fun a0)
   (cond
      ((not (procedure? fun))
       (error "eval" "Not a procedure" name))
      ((not (correct-arity? fun 1))
       (error "eval" "Wrong number of argument" name))
      (else
       (%funcall-1 fun a0))))

;*---------------------------------------------------------------------*/
;*    funcall-2 ...                                                    */
;*---------------------------------------------------------------------*/
(define-inline (funcall-2 name fun a0 a1)
   (cond
      ((not (procedure? fun))
       (error "eval" "Not a procedure" name))
      ((not (correct-arity? fun 2))
       (error "eval" "Wrong number of argument" name))
      (else
       (%funcall-2 fun a0 a1))))

;*---------------------------------------------------------------------*/
;*    funcall-3 ...                                                    */
;*---------------------------------------------------------------------*/
(define-inline (funcall-3 name fun a0 a1 a2)
   (cond
      ((not (procedure? fun))
       (error "eval" "Not a procedure" name))
      ((not (correct-arity? fun 3))
       (error "eval" "Wrong number of argument" name))
      (else
       (%funcall-3 fun a0 a1 a2))))

;*---------------------------------------------------------------------*/
;*    funcall-4 ...                                                    */
;*---------------------------------------------------------------------*/
(define-inline (funcall-4 name fun a0 a1 a2 a3)
   (cond
      ((not (procedure? fun))
       (error "eval" "Not a procedure" name))
      ((not (correct-arity? fun 4))
       (error "eval" "Wrong number of argument" name))
      (else
       (%funcall-4 fun a0 a1 a2 a3))))

;*---------------------------------------------------------------------*/
;*    eval-apply ...                                                   */
;*---------------------------------------------------------------------*/
(define-inline (eval-apply name fun len args)
   (cond
      ((not (procedure? fun))
       (error "apply" "Not a procedure" name))
      ((not (correct-arity? fun len))
       (error "eval" "Wrong number of argument" name))
      (else
       (%eval-apply fun args))))
       
;*---------------------------------------------------------------------*/
;*    Les environments ...                                             */
;*---------------------------------------------------------------------*/
(init-the-global-environment!)
       

